==> makeen-expressionengine-0.0.51.tgz: package/src/functions/globals/groupBy.ts <==
﻿import { List } from 'typescript-dotnet-commonjs/System/Collections/List'
import expressionData from '../../expressionData'
import { expressionPart } from '../../compiler/parts/expressionPart'
import { operationPart } from '../../compiler/parts/operationPart'
import { expressionCompiler } from '../../compiler/expressionCompiler'
import { invalidParametersException } from '../../error/invalidParametersException'
import { mfunction } from '../mfunction'
import { groupDataContext } from '../groupDataContext'
import { aggregateFunction } from '../aggregateFunction'
import { compileFlags } from '../../compiler/expressionCompiler'
import { StringBuilder } from 'typescript-dotnet-commonjs/System/Text/StringBuilder'
import { namedSelection } from '../namedSelection'
import { aggregateFunctionCallPart } from '../../compiler/parts/aggregateFunctionCallPart'
import { outputVariables } from '../../utils/outputVariables'
import { selection } from '../selection'
import { DictionaryObject } from '../../utils/dictionaryObject'

export class groupBy extends mfunction {

    public /* override */ get minParameters(): number {
        return 1;
    }

    public /* override */ get maxParameters(): number {
        return -1;
    }

    public namedSelection: boolean = false;

    public /* internal */ /* override */ compileParameters(paramList: List<string>, flags: compileFlags): List<expressionPart> {
        if ((((this.minParameters > 0)
            && (paramList.count < this.minParameters))
            || ((this.maxParameters >= 0)
                && (paramList.count > this.maxParameters)))) {
            throw new invalidParametersException(super.getFunctionName(), this.minParameters, this.maxParameters);
        }

        let parameters: List<expressionPart> = new List<expressionPart>();
        for (let item of paramList.toArray()) {
            let compileF = new compileFlags(flags);
            compileF.aggregationSection = true;
            parameters.add(expressionCompiler.compile(item, 0, (item.length - 1), compileF) as expressionPart);
        }

        return parameters;
    }

    public invoke(expressionData: expressionData, parameters: List<expressionPart>, /* out */resolved: outputVariables, running: groupDataContext | undefined): Object | undefined {
        let obj: Object = expressionData.contexts.peek();
        resolved.resolved = true;
        let tempData: expressionData = expressionData.toSubEx();
        let isNamedSelection: boolean = (this.namedSelection || parameters.any(p => p instanceof operationPart && (p as operationPart).operator?.name == "@"));
        if ((obj instanceof List)) {
            let result: List<Object> = new List<Object>();
            let done: boolean = false;
            while (!done) {
                result.clear();
                //  Clear();
                try {
                    let groupedFields: DictionaryObject = new DictionaryObject();
                    let groupData: DictionaryObject = new DictionaryObject();
                    for (let item of obj.toArray()) {
                        let itemKeyBuilder: StringBuilder = new StringBuilder();
                        let itemVals: Object[] = new Array(parameters.count);
                        tempData.contexts.push(item);
                        for (let i: number = 0; (i < parameters.count); i++) {
                            let groupExpression: expressionPart = parameters.get(i);
                            if (!this.HasAggregateFunction(groupExpression)) {
                                let subResolved: outputVariables = new outputVariables();
                                let val: Object = groupExpression.resolve(tempData, /* out */subResolved, running) as Object;
                                let groupKey: Object | undefined = undefined;
                                if ((isNamedSelection && !(val instanceof DictionaryObject))) {
                                    let tempVal = val;
                                    val = new DictionaryObject();

                                    (val as DictionaryObject).add(groupExpression.simpleName, tempVal);
                                }

                                if ((val != undefined && val instanceof DictionaryObject)) {
                                    groupKey = (<DictionaryObject>(val)).getValues()[0];
                                }
                                else {
                                    groupKey = val;
                                }

                                itemVals[i] = val;
                                if (subResolved.resolved) {
                                    if (groupKey instanceof DictionaryObject)
                                        itemKeyBuilder.append((groupKey as DictionaryObject).getValues()[0]).append(",");
                                    else
                                        itemKeyBuilder.append(groupKey).append(",");
                                }

                            }

                        }

                        let itemKey: string = itemKeyBuilder.toString();
                        if ((itemKey.length > 0)) {
                            if (!groupedFields.containsKey(itemKey)) {
                                groupedFields[itemKey] = itemVals;
                            }
                            else {
                                itemVals = groupedFields[itemKey];
                            }

                            let itemGroupData: groupDataContext;
                            if (groupData.containsKey(itemKey)) {
                                itemGroupData = groupData.getValue(itemKey) as groupDataContext;
                            }
                            else {
                                itemGroupData = new groupDataContext();
                                groupData.setValue(itemKey, itemGroupData);
                            }

                            for (let i: number = 0; (i < parameters.count); i++) {
                                let groupExpression: expressionPart = parameters.get(i);
                                let subResolved: outputVariables = new outputVariables();
                                if (this.HasAggregateFunction(groupExpression)) {
                                    //  AggregateFunctionCallPart aggregateFunc = groupExpression as AggregateFunctionCallPart;
                                    //  string groupToken = groupExpression.Resolve<string>(tempData, out subResolved);
                                    let val: Object = groupExpression.resolve(tempData, /* out */subResolved, /* running:*/ itemGroupData) as Object;
                                    if (this.namedSelection && !(val instanceof DictionaryObject)) {
                                        val = new DictionaryObject();
                                        (val as DictionaryObject).add(groupExpression.simpleName, val)
                                    }

                                    itemVals[i] = val;
                                }
                            }
                        }

                        tempData.contexts.pop();
                    }

                    for (let values of groupedFields.getValues()) {
                        if (isNamedSelection) {
                            result.add(new namedSelection(values));
                        }
                        else if ((values.length == 1)) {
                            result.add(values[0]);
                        }
                        else {
                            result.add(new selection(values));
                        }
                    }

                    done = true;
                }
                catch (e) {
                    let d = e;
                }

            }

            return result;
        }

        resolved.resolved = false;
        return obj;
    }

    private HasAggregateFunction(part: expressionPart): boolean {
        if ((part instanceof aggregateFunctionCallPart)) {
            return true;
        }
        else if ((part.subParts != null)) {
            for (let item of part.subParts.toArray()) {
                if (this.HasAggregateFunction(item)) {
                    return true;
                }

            }

        }

        return false;
    }
}

==> makeen-expressionengine-0.0.63.tgz: package/src/functions/globals/groupBy.ts <==
﻿import { List } from 'typescript-dotnet-commonjs/System/Collections/List'
import expressionData from '../../expressionData'
import { expressionPart } from '../../compiler/parts/expressionPart'
import { operationPart } from '../../compiler/parts/operationPart'
import { expressionCompiler } from '../../compiler/expressionCompiler'
import { invalidParametersException } from '../../error/invalidParametersException'
import { mfunction } from '../mfunction'
import { groupDataContext } from '../groupDataContext'
import { compileFlags } from '../../compiler/expressionCompiler'
import { StringBuilder } from 'typescript-dotnet-commonjs/System/Text/StringBuilder'
import { namedSelection } from '../namedSelection'
import { aggregateFunctionCallPart } from '../../compiler/parts/aggregateFunctionCallPart'
import { outputVariables } from '../../utils/outputVariables'
import { selection } from '../selection'
import { DictionaryObject } from '../../utils/dictionaryObject'

export class groupBy extends mfunction {

     /* override */ get minParameters(): number {
        return 1;
    }

     /* override */ get maxParameters(): number {
        return -1;
    }

     namedSelection: boolean = false;

     /* internal */ /* override */ compileParameters(paramList: List<string>, flags: compileFlags): expressionPart[] {
        if ((((this.minParameters > 0)
            && (paramList.count < this.minParameters))
            || ((this.maxParameters >= 0)
                && (paramList.count > this.maxParameters)))) {
            throw new invalidParametersException(super.getFunctionName(), this.minParameters, this.maxParameters);
        }

        let parameters: expressionPart[] = [];
        for (let item of paramList.toArray()) {
            let compileF = new compileFlags(flags);
            compileF.aggregationSection = true;
            parameters.push(expressionCompiler.compile(item, 0, (item.length - 1), compileF) as expressionPart);
        }

        return parameters;
    }

     invoke(expressionData: expressionData, parameters: expressionPart[], /* out */resolved: outputVariables, running: groupDataContext | undefined): Object | undefined {
        let obj: Object = expressionData.contexts.peek();
        resolved.resolved = true;
        let tempData: expressionData = expressionData.toSubEx();
        let isNamedSelection: boolean = (this.namedSelection || parameters.some((p) => p instanceof operationPart && (p as operationPart).operator?.name == "@"));
        if ((obj instanceof List)) {
            let result: List<Object> = new List<Object>();
            let done: boolean = false;
            while (!done) {
                result.clear();
                //  Clear();
                try {
                    let groupedFields: DictionaryObject = new DictionaryObject();
                    let groupData: DictionaryObject = new DictionaryObject();
                    for (let item of obj.toArray()) {
                        let itemKeyBuilder: StringBuilder = new StringBuilder();
                        let itemVals: Object[] = new Array(parameters.length);
                        tempData.contexts.push(item);
                        for (let i: number = 0; (i < parameters.length); i++) {
                            let groupExpression: expressionPart = parameters[i];
                            if (!this.HasAggregateFunction(groupExpression)) {
                                let subResolved: outputVariables = new outputVariables();
                                let val: Object = groupExpression.resolve(tempData, /* out */subResolved, running) as Object;
                                let groupKey: Object | undefined = undefined;
                                if ((isNamedSelection && !(val instanceof DictionaryObject))) {
                                    let tempVal = val;
                                    val = new DictionaryObject();

                                    (val as DictionaryObject).add(groupExpression.simpleName, tempVal);
                                }

                                if ((val != undefined && val instanceof DictionaryObject)) {
                                    groupKey = (<DictionaryObject>(val)).getValues()[0];
                                }
                                else {
                                    groupKey = val;
                                }

                                itemVals[i] = val;
                                if (subResolved.resolved) {
                                    if (groupKey instanceof DictionaryObject)
                                        itemKeyBuilder.append((groupKey as DictionaryObject).getValues()[0]).append(",");
                                    else
                                        itemKeyBuilder.append(groupKey).append(",");
                                }

                            }

                        }

                        let itemKey: string = itemKeyBuilder.toString();
                        if ((itemKey.length > 0)) {
                            if (!groupedFields.containsKey(itemKey)) {
                                groupedFields[itemKey] = itemVals;
                            }
                            else {
                                itemVals = groupedFields[itemKey];
                            }

                            let itemGroupData: groupDataContext;
                            if (groupData.containsKey(itemKey)) {
                                itemGroupData = groupData.getValue(itemKey) as groupDataContext;
                            }
                            else {
                                itemGroupData = new groupDataContext();
                                groupData.setValue(itemKey, itemGroupData);
                            }

                            for (let i: number = 0; (i < parameters.length); i++) {
                                let groupExpression: expressionPart = parameters[i];
                                let subResolved: outputVariables = new outputVariables();
                                if (this.HasAggregateFunction(groupExpression)) {
                                    //  AggregateFunctionCallPart aggregateFunc = groupExpression as AggregateFunctionCallPart;
                                    //  string groupToken = groupExpression.Resolve<string>(tempData, out subResolved);
                                    let val: Object = groupExpression.resolve(tempData, /* out */subResolved, /* running:*/ itemGroupData) as Object;
                                    if (this.namedSelection && !(val instanceof DictionaryObject)) {
                                        val = new DictionaryObject();
                                        (val as DictionaryObject).add(groupExpression.simpleName, val)
                                    }

                                    itemVals[i] = val;
                                }
                            }
                        }

                        tempData.contexts.pop();
                    }

                    for (let values of groupedFields.getValues()) {
                        if (isNamedSelection) {
                            result.add(new namedSelection(values));
                        }
                        else if ((values.length == 1)) {
                            result.add(values[0]);
                        }
                        else {
                            result.add(new selection(values));
                        }
                    }

                    done = true;
                }
                catch (e) {
                    throw e;
                }

            }

            return result;
        }

        resolved.resolved = false;
        return obj;
    }

    private HasAggregateFunction(part: expressionPart): boolean {
        if ((part instanceof aggregateFunctionCallPart)) {
            return true;
        }
        else if ((part.subParts != null)) {
            for (let item of part.subParts) {
                if (this.HasAggregateFunction(item)) {
                    return true;
                }

            }

        }

        return false;
    }
}
